﻿#include "Connected.h"

extern int verbose;

set<int> extractFromStack(vector<pair<int, int>>& st, int u, int v) {
	//Extracts all edges in the stack up to and including (u,v). These edges constitute a biconnected component.
	//In this case we save the vertices of the biconnected component.
	set<int> X;
	while (st.back() != pair<int, int>({ u, v })) {
		X.insert(st.back().first);
		X.insert(st.back().second);
		st.pop_back();
	}
	X.insert(st.back().first);
	X.insert(st.back().second);
	st.pop_back();
	return(X);
}
void DFSForBCCs(int u, Graph& G, vector<int>& low, vector<int>& disc, vector<int>& parent, vector<bool>& visited, int& time, vector<pair<int, int>>& st, vector<set<int>>& BCCs) {
	//Conducts a DFS traversal that allows us to identify the biconnected components of graph G. 
	//This is a recursive algorithm -- as a result, for very large graphs we could get a stack overflow error. These
	//are not usually picked up with a try... catch... condition in c++ and can therefore cause an unexplained crash. If this 
	//is the case there are two solutions:
	//(a) Change the compiler options to allow a bigger stack in memory. In Visual Studio, the default stack size is a tiny 1MB.
	//    This can be increased manually by going to Project >> Properties >> Linker >> All Options, and then specifying a different value for 
	//    "Stack Commit Size" and "Stack Reserve Size". This value is specified in bytes, so 10,485,760 specifies 10 MB.
	//(b) A second, more sustainable, option it to implement an iterative version of this algorithm (pending...)
	time++;
	disc[u] = low[u] = time;
	visited[u] = true;
	int v, w, child = 0;
	for (auto& el : G.AList[u]) {
		v = el.vertex;
		w = el.weight;
		if (w != infty) {
			if (!visited[v]) {
				child++;
				st.push_back({ u, v });
				parent[v] = u;
				DFSForBCCs(v, G, low, disc, parent, visited, time, st, BCCs);
				low[u] = minVal(low[u], low[v]);
				if (parent[u] == null && child > 1) {
					//Found new biconnected component. It is found by extracting information from the stack
					BCCs.push_back(extractFromStack(st, u, v));	
				}
				if (parent[u] != null && low[v] >= disc[u]) {
					//Found new biconnected component. It is found by extracting information from the stack
					BCCs.push_back(extractFromStack(st, u, v));
				}
			}
			else if (parent[u] != v && disc[v] < low[u]) {
				low[u] = disc[v];
				st.push_back({ u, v });
			}
		}
	}
}
vector<set<int>> getBCCs(Graph& G) {
	//Returns all biconnected components of the graph as a vector of vertex sets. Biconnected components also include dyads, but not singletons.
	int time = 0;
	vector<bool> visited(G.n, false);
	vector<int> low(G.n), disc(G.n), parent(G.n, null);
	vector<pair<int, int>> st;
	vector<set<int>> BCCs;
	for (int u = 0; u < G.n; u++) {
		if (visited[u] == false) {
			DFSForBCCs(u, G, low, disc, parent, visited, time, st, BCCs);
			if (!st.empty()) {
				//Add any remaining bi-connected component to to BCCs
				BCCs.push_back(set<int>());
				while (!st.empty()) {
					BCCs.back().insert(st.back().first);
					BCCs.back().insert(st.back().second);
					st.pop_back();
				}
			}
		}
	}
	return(BCCs);
}