#include "Graph.h"

extern int verbose;

vector<int> alias;

int getArcWeight(Graph& G, int u, int v) {
	//Returns the weight of arc (u, v) in the adjacency list
	auto it = G.AList[u].find({ v, 0 });
	if (it != G.AList[u].end()) return((*it).weight);
	else { logAndExit("Error: Accessing a weight of a non-existent edge. Ending... \n"); return -1; }
}
int deleteArc(Graph& G, int u, int v) {
	//Deletes the arc (u, v) in the adjacency list and returns its weight
	auto it = G.AList[u].find({ v, 0 });
	int w = (*it).weight;
	G.AList[u].erase(it);
	return w;
}
void addArc(Graph& G, int u, int v, int w) {
	//Adds the arc (u, v) with weight w to the adjacency list
	G.AList[u].insert({ v, w });
}
bool existsArc(Graph& G, int u, int v) {
	//Checks if (u,v) exists in the adjacency list
	if (G.AList[u].count({ v, 0 }) != 0) return true;
	else return false;
}
int getNHeaviestWeights(Graph& G) {
	//Calculate the total weight of the n heaviest edges in G
	vector<int> W;
	for (int u = 0; u < G.n; u++) {
		for (auto& el : G.AList[u]) {
			if (el.vertex > u) W.push_back(el.weight);
		}
	}
	int total = 0;
	sort(W.begin(), W.end(), [](const int a, const int b) {return a > b; });
	for (int i = 0; i < G.n; i++) {
		total += W[i];
	}
	return total;
}
int getHeaviestTwoWeightsFromEachVertex(Graph& G) {
	//Calculate the heaviest two weights from each vertex. Only works for BCCs
	int first, second, total = 0;
	for (int u = 0; u < G.n; u++) {
		first = second = 0;
		for (auto& el : G.AList[u]) {
			if (el.weight > first) {
				second = first;
				first = el.weight;
			}
			else if (el.weight == first) {
				second = first;
			}
			else if (el.weight < first && el.weight > second) {
				second = el.weight;
			}
		}
		total += first + second;
	}
	//Upper bound is half of this total
	return total / 2;
}
int getUB(Graph& G) {
	//Calculate two upper bounds on the length of cycles that can be formed and take the best one. 
	return(minVal(getNHeaviestWeights(G), getHeaviestTwoWeightsFromEachVertex(G)));
}
void trimGraph(Graph& G, Sol& C, Sol& BSF, int UB) {
	//Set G to be the subgraph induced by all vertices <= UB/2 distance from the source. 
	vector<int> L, prev;
	set<int> X;
	int u, maxDist = (UB / 2) + 1;
	DijkstraMaxDist(G, G.source, -1, maxDist, L, prev);
	for (u = 0; u < G.n; u++) {
		if (L[u] < maxDist) X.insert(u);
	}
	if (verbose >= 2) cout << "  Trimming. Removing vertices whose distance from the source is >= floor(" << UB << " / 2) + 1 = " << maxDist << ". This means n = " << G.n << ", m = " << G.m << " becomes n = ";
	if (X.size() < G.n) {
		G = getInducedGraph(G, X);
		//Also need to relabel things in the solutions
		for (u = 0; u < C.seq.size(); u++) {
			C.seq[u] = alias[C.seq[u]];
		}
		for (u = 0; u < BSF.seq.size(); u++) {
			BSF.seq[u] = alias[BSF.seq[u]];
		}
	}
	if (verbose >= 2) cout << G.n << ", m=" << G.m << "\n";
}
Graph getInducedGraph(Graph& G, set<int>& X) {
	//Return the induced subgraph of G specified by the vertices in X. Start by making an alias for each vertex we will keep
	alias.clear();
	alias.resize(G.n, null);
	int cnt = 0, x;
	for (auto& v : X) {
		alias[v] = cnt++;
	}
	//Now construct the new graph GPrime from G
	Graph GPrime;
	GPrime.n = X.size();
	GPrime.AList.resize(GPrime.n, set<Neighbour, minNeighbour>());
	GPrime.vertexID.resize(GPrime.n);
	GPrime.AVec.resize(G.n, vector<Neighbour>());
	for (auto& u : X) {
		x = alias[u];
		GPrime.vertexID[x] = G.vertexID[u];
		for (auto& v : G.AList[u]) {
			if (alias[v.vertex] != null) {
				addArc(GPrime, x, alias[v.vertex], v.weight);
				GPrime.m++;
			}
		}
	}
	GPrime.m /= 2;
	//If we're seeking the longest cycle, set k to an upper bound on cycle length
	GPrime.LenUB = getUB(GPrime);
	GPrime.k = minVal(G.k, GPrime.LenUB);
	//if appropriate, also relabel the source
	if(G.source >=0) GPrime.source = alias[G.source];
	return GPrime;
}
Graph makeGraph(vector<set<Neighbour, minNeighbour> >& adjList, vector<string>& vName, int targetK, int startNode) {
	//Construct the graph G from a vector of vertex names, an adjacency list, a target k, and a start node
	Graph G;
	G.source = startNode;
	G.n = adjList.size();
	G.AList.swap(adjList);
	G.vertexID.swap(vName);
	G.AVec.resize(G.n, vector<Neighbour>());
	for (int u = 0; u < G.n; u++) {
		G.m += G.AList[u].size();
	}
	G.m /= 2;
	//If we're seeking the longest cycle, set k to an upper bound on cycle length
	G.LenUB = getNHeaviestWeights(G);
	G.k = minVal(targetK, G.LenUB);
	if (verbose >= 1) {
		if (verbose >= 2) printGraph(G);
		cout << "Initial graph has " << G.n << " vertices, " << G.m << " edges. Target k = " << G.k << endl;
		if (G.k == G.LenUB) cout << "This k-value is an upper bound on cycle length. Hence the longest cycle is being sought\n";
	}
	return(G);
}