#include "GraphStats.h"
#include "Connected.h"
#include "Graph.h"

void printGraphStats(GraphStats& X) {
	cout << "\tNumVertices-n:     " << X.n << endl;
	cout << "\tNumEdges-m:        " << X.m << endl;
	cout << "\tDesired-Len:       " << X.k << endl;
	cout << "\tMinEdgeWeight:     " << X.minWeight << endl;
	cout << "\tMeanEdgeWeight:    " << X.meanWeight << endl;
	cout << "\tMaxEdgeWeight:     " << X.maxWeight << endl;
	cout << "\tMinDegree:         " << X.minDeg << endl;
	cout << "\tMinWeightedDegree: " << X.minWeightedDeg << endl;
	cout << "\tMaxDegree:         " << X.maxDeg << endl;
	cout << "\tMaxWeightedDegree: " << X.maxWeightedDeg << endl;
	cout << "\tShortestCycle:     [";
	for (auto& v : X.shortestCycle.seq) cout << " " << v;
	cout << " ]\n";
	cout << "\tLB(WeightedGirth): " << X.LB1 << " (The shortest (s-)cycle in G has this length.)" << endl;
	cout << "\tLB(BondyAndFan):   " << X.LB2 << " (= ceil((2 x w(G)) / (n-1)). Must be a cycle of at least this length)" << endl;
	cout << "\tLB(MinDeg):        " << X.LB3 << " (either (a) for all vertices v, there exists a non-Hamiltonian v-cycle of length >= this value; OR (b) all longest cycles in G are Hamiltonian)" << endl;
	cout << "\tLB(MinDegPair):    " << X.LB4 << " (either (a) for all vertices v, there exists a v-cycle of length >= this value; OR (b) G is Hamiltonian)" << endl;
	cout << "\tUB:                " << X.UB1 << " (Sum of all edge weights)" << endl;
	cout << "\tUB:                " << X.UB2 << " (Sum of n heaviest edges)" << endl;
	cout << "\tUB:                " << X.UB3 << " (Two heaviest weights adjacent to a verex, summed across all vertices, divided by two)" << endl;
}
Sol getShortestCycle(Graph& G) {
	//Calculates the the weighted girth of a graph. This can either be the shortest cycle, or the shortest s-cycle.
	int u, v, minu = -1, minv = -1, w, minW = infty;
	vector<int> prev, L;
	set<WArc, minWArcByWeight> E;
	if (G.source < 0) {
		//Make a set of all edges in G, ordered by ascending weight	
		for (u = 0; u < G.n; u++) {
			for (auto& el : G.AList[u]) {
				if (el.vertex > u) E.emplace(WArc{ u, el.vertex, el.weight });
			}
		}
	}
	else {
		//Make a set of all edges adjacent to to the source, ordered by ascending weight	
		for (auto& el : G.AList[G.source]) {
			E.emplace(WArc{ G.source, el.vertex, el.weight });
		}
	}
	//Remove an edge {u,v} then get the sortest u-v path. This gives us the shortest cycle involving this edge
	for (auto& el : E) {
		u = el.u;
		v = el.v;
		w = el.weight;
		if (w >= minW) {
			//Removal of all remaining edges {u,v} will give cycles longer than the best observed
			break;
		}
		else {
			//Remove edge {u, v}
			deleteArc(G, u, v);
			deleteArc(G, v, u);
			//Identify the shortest u-v-path using Dijkstra's and see it the resultant cycle is the best so far
			DijkstraMaxDist(G, u, v, minW - w, L, prev);
			if (L[v] < infty && L[v] + w < minW) {
				minW = L[v] + w;
				minu = u;
				minv = v;
			}
			//Reinstate edge {s, v} and end
			addArc(G, u, v, w);
			addArc(G, v, u, w);
		}
	}
	//Now determine the shortest cycle in the graph 
	w = getArcWeight(G, minu, minv);
	deleteArc(G, minu, minv);
	deleteArc(G, minv, minu);
	DijkstraTarget(G, minu, minv, L, prev);
	Sol C = getPathFromPrev(minu, minv, L, prev);
	C.seq.push_back(minu);
	C.len = minW;
	//Reinstate edge {u, v} and end
	addArc(G, minu, minv, w);
	addArc(G, minv, minu, w);
	return C;
}
GraphStats getGraphStats(Graph& G) {
	//Parse the graph G (which must be a biconnected component) and produce a number of stats. First, calculate the maximum and total weights
	int u, v;
	GraphStats X;
	vector<int> weightedDeg(G.n, 0);
	X.m = G.m;
	X.n = G.n;
	X.k = G.k;
	X.source = G.source;
	for (u = 0; u < G.n; u++) {
		for (auto& el : G.AList[u]) {
			weightedDeg[u] += el.weight;
			if (el.weight > X.maxWeight) X.maxWeight = el.weight;
			if (el.weight < X.minWeight) X.minWeight = el.weight;
		}
	}
	X.meanWeight = X.totalWeight / double(G.m);
	
	//Identify the shortest cycle in G
	X.shortestCycle = getShortestCycle(G);
	X.LB1 = X.shortestCycle.len;

	//Get Bondy-Fan Bound
	X.LB2 = (int)ceil((2.0 * X.totalWeight) / ((double)G.n - 1));

	//Derive info about the weighted degrees
	for (u = 0; u < G.n; u++) {
		if (G.AList[u].size() < X.minDeg) X.minDeg = G.AList[u].size();
		if (weightedDeg[u] < X.minWeightedDeg) X.minWeightedDeg = weightedDeg[u];
		if (G.AList[u].size() > X.maxDeg) X.maxDeg = G.AList[u].size();
		if (weightedDeg[u] > X.maxWeightedDeg) X.maxWeightedDeg = weightedDeg[u];
	}
	X.LB3 = 2 * X.minWeightedDeg;

	//Get minimum weighted vertex sum of nonadjacent vertices
	int minDegPair = infty;
	for (u = 0; u < G.n - 1; u++) {
		for (v = u + 1; v < G.n; v++) {
			if (!existsArc(G, u, v) && weightedDeg[u] + weightedDeg[v] < minDegPair) {
				minDegPair = weightedDeg[u] + weightedDeg[v];
			}
		}
	}
	X.LB4 = minDegPair;

	//Return three upper bounds (a) UB1 = sum of all edge weights, (b) UB2 = sum of n heaviest edges, (c) UB3 = sum of the heaviest 2 edges on each vertex divided by two
	X.UB1 = 0;
	for (u = 0; u < G.n; u++) {
		for (auto& el : G.AList[u]) {
			X.UB1 += el.weight;
		}
	}
	X.UB1 /= 2;
	X.UB2 = getNHeaviestWeights(G);
	X.UB3 = getHeaviestTwoWeightsFromEachVertex(G);
	return X;
}