#include "IO.h"

extern int verbose;

void readInputFile(string fname, vector<set<Neighbour, minNeighbour> >& adjList, vector<string>& vName) {
	//Reads an edge-weighted undirected graph in DIMACS format file and creates the corresponding adjacency list. All weights must be nonnegative
	ifstream inStream;
	inStream.open(fname);
	if (inStream.fail()) {
		logAndExit("Error opening input file " + fname + ". Exiting...\n");
	}
	if(verbose >= 1) cout << "Input = " << fname << endl;
	int line = 1, i, j, n, m;
	long long w, totalW=0;
	set<set<int>> E;
	string str, name;
	try {
		while (true) {
			inStream >> str;
			if (str == "c") {
				//Ignore the rest of the line, it is a comment
				getline(inStream, str);
				line++;
			}
			else if (str == "p") {
				//Read the basic information about the graph
				inStream >> n >> m;
				line++;
				break;
			}
			else {
				logAndExit("Invalid input in file <" + fname + ">. Exiting...\n");
			}
		}
		//Set up the data structures
		vName.resize(n, "");
		adjList.resize(n, set<Neighbour, minNeighbour>());
		//Now read in all of the vertices' aliases
		if (verbose >= 1) {
			cout << "Reading " << n << " vertex names ... \n";
		}
		for (int v = 0; v < n; v++) {
			//Read in information about a vertex 
			inStream >> str >> name;
			if (str != "v") {
				logAndExit("Error. Corruption on line " + to_string(line) + " when reading vertex IDs. Exiting...\n");
			}
			vName[v] = name;
			line++;
		}
		//Now read in all of the edges
		if (verbose >= 1) {
			cout << "Reading " << m << " edges ... \n";
		}
		for (int e = 0; e < m; e++) {
			//Read in information about an edge: its start point i, endpoint j, and weight w
			inStream >> str >> i >> j >> w;
			if (str != "e" || i < 0 || i >= n || j < 0 || j >= n) {
				logAndExit("Error. Corruption on line " + to_string(line) + " when reading an edge. Exiting...\n");
			}
			//We are looking to add edge {i,j} with weight w. Do a check for duplicates and out of bounds
			if (w >= (INT_MAX/2) || w < 0) {
				logAndExit("Error. Corruption on line " + to_string(line) + ". Weights in the input file should be in the set {0, 1,..., " + to_string(INT_MAX / 2) + "}. Exiting...\n");
			}
			totalW += w;
			if (i == j) {
				cout << "Warning from line " << to_string(line) << ". Edge {" + to_string(i) + ", " + to_string(j) + "} is a loop and will be ignored\n";
			}
			else if (E.count({ i,j }) > 0 ) {
				cout << "Warning from line " << to_string(line) << ". Edge {" + to_string(i) + ", " + to_string(j) + "} is already defined and is therefore ignored\n";
			}
			else {
				//Assuming it is OK, the information is added to the adjacency list
				adjList[i].insert({ j, (int)w });
				adjList[j].insert({ i, (int)w });
				E.insert({ i,j });
			}
			line++;
		}
	}
	catch (...) {
		logAndExit("Error: Corrupted input file (line " + to_string(line) + "). Exiting...\n");
	}
	inStream.close();
	if (totalW >= INT_MAX / 2) {
		logAndExit("Error. Total of all edge weights in this file is gretater than " + to_string(INT_MAX / 2) + ". Risk of overflow. Exiting...\n");
	}
}
void printGraph(Graph& G) {
	//Prints a graph as an adjacency list. Can cause a lot output
	cout << "Considered graph is as follows:\nn = " << G.n << "\nm = " << G.m << "\nsource = " << G.source << "\n";
	cout << setw(8) << "index" << setw(12) << "vertexID" << "\tNeighbours (vertex, weight)\n";
	int u, v, w;
	for (u = 0; u < G.n; u++) {
		cout << setw(8) << u << setw(12) << G.vertexID[u] << "\t{ ";
		for (auto& el : G.AList[u]) {
			v = el.vertex;
			w = el.weight;
			if (w != infty)	cout << "(" << v << "," << w << ") ";
			else			cout << "(" << v << ",inf.) ";
		}
		cout << "}\n";
	}
	cout << "--------------\n\n";
}
int getSolWeight(Graph& G, Sol& C) {
	//Goes through a sequence of vertices and returns it length
	if (C.seq.size() <= 1) return 0;
	int tot = 0;
	for (int i = 0; i < C.seq.size() - 1; i++) {
		tot += getArcWeight(G, C.seq[i], C.seq[i + 1]);
	}
	return tot;
}
string prettySolStringOrigLabels(Graph& G, Sol& C) {
	if (C.len == infty || C.seq.size() <= 1) {
		return("[-]");
	}
	string str = "[";
	for (int i = 0; i < C.seq.size() - 1; i++) {
		str += G.vertexID[C.seq[i]] + ",";
	}
	str += G.vertexID[C.seq.back()] + "]";
	return str;
}
string prettySolString(Graph& G, Sol& C) {
	if (C.len == infty || C.seq.size() <= 1) {
		return("[-]");
	}
	string str = "[";
	for (int i = 0; i < C.seq.size() - 1; i++) {
		str += to_string(C.seq[i]) + ",";
	}
	str += to_string(C.seq.back()) + "]";
	return str;
}
void printSolAndCheck(Graph& G, Sol &C, bool doOutput) {
	//doOutput is set to true by defauk. If this is the case this prints a cycle as a sequence of vertices. 
	//It also checks the claimed weight C.len is correct, plus a few other checks
	if (doOutput) {
		if (C.len == infty) {
			cout << setw(10) << "inf." << setw(10) << "inf." << setw(8) << "0" << "\t" << prettySolString(G, C) << "\n";
			return;
		}
		if (C.seq.size() <= 1) {
			cout << setw(10) << G.k << setw(10) << "0" << setw(8) << "0" << "\t" << prettySolString(G, C) << "\n";
			return;
		}
		cout << setw(10) << abs(C.len - G.k) << setw(10) << C.len << setw(8) << C.seq.size() - 1 << "\t" << prettySolString(G, C) << "\n";;
	}
	string str = "";
	if (getSolWeight(G, C) != C.len) {
		str += "The cycle's weight is not stored correctly. It should be " + to_string(getSolWeight(G, C)) + "\n";
	}
	//Check that the source is at both ends
	if(C.seq.front() != C.seq.back())
		str += "Error: The same vertex should appear at the front and back of this vertex sequence.\n";
	//Check for repeated vertices
	vector<int> freq(G.n, 0);
	for (auto& v : C.seq) freq[v]++;
	freq[C.seq.front()]--;
	for (int v = 0; v < G.n; v++) {
		if (freq[v] > 1) {
			str += "Error: vertex v-" + to_string(v) + " (ID:" + G.vertexID[v] + ") is repeated in the cycle\n";
		}
	}
	//Check for repeated edges
	if (str != "") {
		cout << prettySolString(G, C) << endl;
		logAndExit(str);
	}
	
}
string makePythonOSMString(Graph& G, Sol& C) {
	//Go through a cycle, convert node labels to their OSM labels, eliminate consecutive duplicates, and put in a string representing a Python list
	int i;
	if (C.seq.empty()) return("[]");
	string str = "[";
	for (i = 0; i < C.seq.size() - 1; i++) {
		str += G.vertexID[C.seq[i]] + ", ";
	}
	str += G.vertexID[C.seq[i]] + "]";
	return str;
}
void writeSolToFile(Graph& G, Sol& C, string& fName) {
	//Prints out the vector of cycles as a list of lists (Python)
	ofstream fout(fName + "_sol.txt");
	//Print out each cycle. 
	fout << "[\n";
	fout << " " << makePythonOSMString(G, C) << ",\n";
	fout << "]\n";
	fout.close();
}