#include "LocalSearch.h"

extern int verbose;

vector<int> Perm, NewSeq, PathVec;
set<int> PathSet;

void getPaths(Graph& G, int s, set<int>& T, vector<int>& L, vector<int>& prev, int SPAlg) {
	//Decides which spanning tree algorithm to use
	switch (SPAlg) {
	case 1:
		RFSSet(G, s, T, L, prev);
		break;
	case 2:
		randomBFSSet(G, s, T, L, prev); 
		break;
	case 3:
		randomDFSSet(G, s, T, L, prev);
		break;
	case 4:
		WilsonSet(G, s, T, L, prev);
		break;
	default:
		DijkstraSet(G, s, T, L, prev);
		break;
	}
}
Sol getInitSol(Graph& G, int SPAlg) {
	//If source vertex s is undefined, choose one at random
	int s = G.source;
	if (s < 0) {
		do {
			s = rand() % G.n;
		} while (G.AList[s].empty());
	}
	//Now select a random vertex v adjacent to s
	int i = rand() % G.AList[s].size();
	auto it = G.AList[s].begin();
	advance(it, i);
	int v = (*it).vertex;
	int w = (*it).weight;
	//Remove edge {s, v}
	deleteArc(G, s, v);
	deleteArc(G, v, s);
	//Identify an s-v-path
	vector<int> prev, L;
	set<int> T = { v };
	getPaths(G, s, T, L, prev, SPAlg);
	Sol C = getPathFromPrev(s, v, L, prev);
	C.seq.push_back(s);
	C.len += w;
	//Reinstate edge {s, v} and end
	addArc(G, s, v, w);
	addArc(G, v, s, w);
	return C;
}
void getPerm(Graph& G, Sol& C, bool doingSCycles) {
	//Set global variable Perm as a random permutation of the values 0,1,...,n-1. These refer to the positions of the vertices in C
	Perm.clear();
	for (int i = 0; i < C.seq.size(); i++) Perm.push_back(i);
	//Also include a "null" in the permutation if we are no considering s-cycles
	if (!doingSCycles) Perm.push_back(-1);
	random_shuffle(Perm.begin(), Perm.end());
}
vector<int> getCumulative(Graph& G, vector<int>& C) {
	//Produce vector of the cumulative length of the cycle C
	if (C.size() <= 2) return(vector<int>{});
	vector<int> A = { 0 };
	for (int i = 1; i < C.size(); i++) {
		A.push_back(A[i - 1] + getArcWeight(G, C[i - 1], C[i]));
	}
	return(A);
}
void getPathSet(int s, int t, vector<int>& prev, vector<int>& L) {
	//Populates the global variable PathSet so that it holds all vertices in the s-t-path as a set
	PathSet.clear();
	if (L[t] < infty) {
		int u = t;
		while (u != s) {
			PathSet.insert(u);
			u = prev[u];
		}
		PathSet.insert(s);
	}
}
void getPathVec(Graph& G, int s, int t, vector<int>& prev, vector<int>& L, bool flip) {
	//Populates the global variable PathVec so that it holds all vertices in a vector
	//if flip is true, returns a t-s-path, else returns an s-t-path
	PathVec.clear();
	if (L[t] < infty) {
		int u = t;
		while (u != s) {
			PathVec.push_back(u);
			u = prev[u];
		}
		if (PathVec[0] == G.n) PathVec[0] = s;
		PathVec.push_back(s);
		if (!flip) reverse(PathVec.begin(), PathVec.end());
	}
}
bool hasThreePlusEdges(int s, int t, vector<int>& prev) {
	//Conctant-time operation to ensure that the path from s to t in prev has more than two edges
	int count = 0, u = t;
	while (u != s) {
		count++;
		if (count > 3) return true;
		u = prev[u];
	}
	count++;
	if (count > 3) return true;
	else return false;
}
int evaluateMove1(Sol& sol, int uPos, int vPos, int v, vector<int>& L) {
	//assuming that uPos != vPos
	if (uPos < vPos)	return sol.len - (sol.cumulative[vPos] - sol.cumulative[uPos]) + L[v];
	else				return sol.len - (sol.cumulative[uPos] - sol.cumulative[vPos]) + L[v];
}
int evaluateMove2(Sol& sol, int uPos, int vPos, int v, vector<int>& L) {
	//assuming that uPos != vPos
	if (uPos < vPos)	return (sol.cumulative[vPos] - sol.cumulative[uPos]) + L[v];
	else				return (sol.cumulative[uPos] - sol.cumulative[vPos]) + L[v];
}
bool move1IsFeasible(Graph& G, Sol& sol, int uPos, int vPos, vector<int>& prev, vector<int>& L) {
	//Given the sequence with indices [0,...,uPos,...,vPos,...,len-1], we are looking to replace everything between and including [uPos,...,vPos] with a new path.
	//We therefore need to check for common elements between [0,...,uPos-1] and PathSet, and then [vPos+1,...,len-1] and PathSet 
	getPathSet(sol.seq[uPos], sol.seq[vPos], prev, L);
	if (uPos > vPos) {
		swapVals(uPos, vPos);
	}
	else if (uPos == 0) {
		//if this is true then, because the source is at the start and end of the sequence, we need to remove it from PathSet
		PathSet.erase(sol.seq[uPos]);
	}
	for (int i = 0; i < uPos; i++) {
		if (PathSet.count(sol.seq[i]) >= 1) return false;
	}
	for (int i = vPos + 1; i < sol.seq.size(); i++) {
		if (PathSet.count(sol.seq[i]) >= 1) return false;
	}
	return true;
}
bool move2IsFeasible(Graph& G, Sol& sol, int uPos, int vPos, vector<int>& prev, vector<int>& L) {
	//Given the sequence with indices [0,...,uPos,...,vPos,...,len-1], we are looking to replace everything outside of [uPos,...,vPos] with a new path
	//We therefore need to check for common elements between [uPos,...,vPos] and PathSet,
	getPathSet(sol.seq[uPos], sol.seq[vPos], prev, L);
	if (uPos > vPos) swapVals(uPos, vPos);
	for (int i = uPos + 1; i < vPos; i++) {
		if (PathSet.count(sol.seq[i]) >= 1) return false;
	}
	return true;
}
void doMove1(Graph& G, Sol& sol, int uPos, int vPos, vector<int>& P, int newLen) {
	//Given the sequnce [....,uPos,...,vPos,...], replace everything between and including [uPos,...,vPos] with P
	if (uPos > vPos) swapVals(uPos, vPos);
	NewSeq.clear();
	for (int i = 0; i < uPos; i++) NewSeq.push_back(sol.seq[i]);
	for (int i = 0; i < P.size(); i++) NewSeq.push_back(P[i]);
	for (int i = vPos + 1; i < sol.seq.size(); i++) NewSeq.push_back(sol.seq[i]);
	sol.seq.swap(NewSeq);
	sol.len = newLen;
	sol.cumulative = getCumulative(G, sol.seq);
}
void doMove2(Graph& G, Sol& sol, int uPos, int vPos, vector<int>& P, int newLen) {
	//Given the sequnce [0....,uPos,...,vPos,...len-1], replace everything outside of [uPos,...,vPos] with P
	if (uPos > vPos) swapVals(uPos, vPos);
	NewSeq.clear();
	reverse(P.begin(), P.end());
	for (int i = uPos; i <= vPos; i++) NewSeq.push_back(sol.seq[i]);
	for (int i = 1; i < P.size(); i++) NewSeq.push_back(P[i]);
	sol.seq.swap(NewSeq);
	sol.len = newLen;
	sol.cumulative = getCumulative(G, sol.seq);
}
void doMove3(Graph& G, Sol& sol, vector<int>& P, int newLen) {
	//Replaces the contents of C with the vector P
	sol.seq.swap(P);
	sol.len = newLen;
	sol.cumulative = getCumulative(G, sol.seq);
}
void doMove4(Graph& G, Sol& sol, Sol& tempSol) {
	sol = tempSol;
	sol.atLocalOpt = false;
	sol.cumulative = getCumulative(G, sol.seq);
}
void adjustB(double &B, clock_t& MChainStart, Sol& C, Sol& CBest, Graph& G, int reportInterval, int& accCount, int& feasCount, clock_t clockLimit, int& globalUB) {
	//Output details of most recent Markov chain
	if (verbose >= 1) {
		cout << setw(10) << (clockLimit - clock()) / (double)(CLOCKS_PER_SEC)
			<< setw(6) << reportInterval
			<< setw(6) << accCount
			<< setw(6) << feasCount
			<< setw(15) << B
			<< setw(10) << G.k
			<< setw(10) << abs(C.len - G.k)
			<< setw(10) << C.len
			<< setw(10) << abs(CBest.len - G.k)
			<< setw(10) << CBest.len
			<< setw(12) << globalUB << "\n";
		//printSolAndCheck(G, C);
		//printSolAndCheck(G, CBest);
	}
	//Update B for the next "Markov chain"
	double muStar = (double)(clockLimit - clock()) / (double)(clock() - MChainStart);
	double alpha = B / muStar;
	B -= alpha;
	accCount = 0;
	feasCount = 0;
	MChainStart = clock();
}
void modifyGraph(Graph& G, Sol& C, int x, set<WArc, minWArc>& H_E, set<int>& H_V) {
	//This modifies G by first removing all edges of cycle C (by setting all of its edges to infinity weights). The original edge weights are stored in H_E
	//and the vertices of C are stored in H_V
	int i, u, v, w;
	H_E.clear();
	H_V.clear();
	for (i = 0; i < C.seq.size() - 1; i++) {
		u = C.seq[i];
		v = C.seq[i + 1];
		w = deleteArc(G, u, v);
		H_E.insert({ u,v,w });
		w = deleteArc(G, v, u);
		H_E.insert({ v,u,w });
		H_V.insert(u);
	}
	H_V.insert(v);
	for (auto& el : H_E) {
		addArc(G, el.u, el.v, infty);
	}
	//In addition, we also add a clone of vertex x. This is stored as the last vertex in the graph
	int xPrime = G.n;
	G.AList.push_back(set<Neighbour, minNeighbour>());
	G.AVec.push_back(vector<Neighbour>());
	G.n++;
	for (Neighbour el : G.AList[x]) {
		v = el.vertex;
		w = el.weight;
		//Remove edge (v,x) and add edge (v,xPrime,w)
		deleteArc(G, v, x);
		addArc(G, v, xPrime, w);
	}
	H_V.insert(xPrime);
}
void resetGraph(Graph& G, int x, set<WArc, minWArc>& H_E) {
	//First redirect all of xPrime's in-edges to x, and remove xPrime altogether
	G.n--;
	G.AList.pop_back();
	G.AVec.pop_back();
	int xPrime = G.n, v, w;
	for (Neighbour el : G.AList[x]) {
		v = el.vertex;
		w = el.weight;
		//Remove edge (v,xPrime) and add edge (v,x,w)
		deleteArc(G, v, xPrime);
		addArc(G, v, x, w);
	}
	//Now use H_E to reset the edge weights of the cycle in the graph
	for (auto& el : H_E) {
		deleteArc(G, el.u, el.v);
	}
	for (auto& el : H_E) {
		addArc(G, el.u, el.v, el.weight);
	}
}
void doGreatDeluge(Graph& G, Sol& C, int SPAlg, clock_t clockLimit, bool doingSCycles, bool trimmingGraphs, int& globalUB) {
	if (C.seq.size() <= 2) {
		logAndExit("Cannot apply Great Deluge to a null solution ([ ], [s], or [s, s]). Exiting\n");
	}
	//Variables used
	vector<int> prev, P, L;
	set<WArc, minWArc> H_E;
	set<int> H_V;
	int u = -1, v = -1, uPos, vPos, savedMove = null, itCount = 0, reportInterval = 500, accCount = 0, feasCount = 0, newLen, improvement;
	double B = double(abs(C.len - G.k) * 1.10);
	bool toFlip;
	//Add values to the remaining fields in the Sol struct
	C.atLocalOpt = false;
	C.cumulative = getCumulative(G, C.seq);
	Sol CBest = C, CTemp;
	if (verbose >= 1) {
		if (verbose >= 1) {
			cout << setw(10) << "TimeLeft"
				<< setw(6) << "MCLen"
				<< setw(6) << "Ac"
				<< setw(6) << "Ac+Fs"
				<< setw(15) << "B"
				<< setw(10) << "k"
				<< setw(10) << "cost(C)"
				<< setw(10) << "len(C)"
				<< setw(10) << "cost(BSF)"
				<< setw(10) << "len(BSF)"
				<< setw(12) << "globalUB" << endl;
			cout << setw(10) << (clockLimit - clock()) / (double)(CLOCKS_PER_SEC)
				<< setw(6) << "-"
				<< setw(6) << "-"
				<< setw(6) << "-"
				<< setw(15) << B
				<< setw(10) << G.k
				<< setw(10) << abs(C.len - G.k)
				<< setw(10) << C.len
				<< setw(10) << abs(CBest.len - G.k)
				<< setw(10) << CBest.len
				<< setw(12) << globalUB << "\n";
		}
	}
	//If desired, trim the subgraphs based on globalUB 
	if (doingSCycles && trimmingGraphs && globalUB < infty) {
		trimGraph(G, C, CBest, globalUB);
		if (G.n == G.m) {
			if (verbose >= 1) cout << "Graph has been trimmed so that n = m. Exiting optimisation process\n";
			return;
		}
	}
	clock_t MChainStart = clock();
	while (true) {
		//Initialise variables for this iteration of GD
		savedMove = null;
		//Chose random vertex u in the cycle
		uPos = rand() % (C.seq.size() - 1);
		u = C.seq[uPos];
		//Remove the edges of C from G and also add u' to G too 
		modifyGraph(G, C, u, H_E, H_V);
		//Now get the paths from u to all vertices in (V(C) + u)
		getPaths(G, u, H_V, L, prev, SPAlg);
		//Reset the graph G br reinstating C and deleting u'
		resetGraph(G, u, H_E);
		//Now go through each vertex in C (in a random order) and assess the implications of altering the u-v-paths
		getPerm(G, C, doingSCycles);
		for (int el : Perm) {
			if (clock() >= clockLimit) break;
			vPos = el;
			if (vPos == -1) {
				//This is the special case where we will make a cycle from a randomly selected vertex v
				itCount++;
				CTemp = getInitSol(G, SPAlg);
				if (itCount % reportInterval == 0) adjustB(B, MChainStart, C, CBest, G, reportInterval, accCount, feasCount, clockLimit, globalUB);
				newLen = CTemp.len;
				if (abs(newLen - G.k) <= abs(C.len - G.k) || abs(newLen - G.k) <= B) {
					//This move has been accepted, so do it now by breaking
					accCount++;
					feasCount++;
					savedMove = 4;
					break;
				}
			}
			else {
				//We are going to evaluate a neighbourhood move involving vertices u and v in C
				v = C.seq[vPos];
				if (v != u && L[v] != infty) {
					//We are evaluating the effects of replacing a u-v path in C with a new u-v-path in C
					itCount++;
					if (itCount % reportInterval == 0) adjustB(B, MChainStart, C, CBest, G, reportInterval, accCount, feasCount, clockLimit, globalUB);
					newLen = evaluateMove1(C, uPos, vPos, v, L);
					if (abs(newLen - G.k) <= abs(C.len - G.k) || abs(newLen - G.k) <= B) {
						accCount++;
						if (move1IsFeasible(G, C, uPos, vPos, prev, L)) {
							//This move has been accepted, so do it now by breaking
							feasCount++;
							savedMove = 1;
							if (uPos <= vPos)	toFlip = false;
							else				toFlip = true;
							break;
						}
					}
					if (!doingSCycles) {
						itCount++;
						if (itCount % reportInterval == 0) adjustB(B, MChainStart, C, CBest, G, reportInterval, accCount, feasCount, clockLimit, globalUB);
						newLen = evaluateMove2(C, uPos, vPos, v, L);
						if (abs(newLen - G.k) <= abs(C.len - G.k) || abs(newLen - G.k) <= B) {
							accCount++;
							if (move2IsFeasible(G, C, uPos, vPos, prev, L)) {
								//This move has been accepted, so do it now by breaking
								feasCount++;
								savedMove = 2;
								if (uPos <= vPos)	toFlip = false;
								else				toFlip = true;
								break;
							}
						}
					}
				}
				else if (v == u && L[G.n] != infty) {
					if (!doingSCycles || u == G.source) {
						//We are evaluating the effects of replacing the whole of C with the u-u'-path defined in P
						itCount++;
						if (itCount % reportInterval == 0) adjustB(B, MChainStart, C, CBest, G, reportInterval, accCount, feasCount, clockLimit, globalUB);
						newLen = L[G.n];
						if (abs(newLen - G.k) <= abs(C.len - G.k) || abs(newLen - G.k) <= B) {
							accCount++;
							if (hasThreePlusEdges(u, G.n, prev)) {
								//This move has been accepted, so do it now by breaking
								feasCount++;
								savedMove = 3;
								break;
							}
						}
					}
				}
			}
		}
		//If we are here we are going to do a move
		if (clock() >= clockLimit) break;
		if (savedMove != null) {
			//A move has been accepted, so do it and check of this gives the bestSoFar
			if (savedMove == 4) {
				doMove4(G, C, CTemp);
			}
			else {
				if (v != u) getPathVec(G, u, v, prev, L, toFlip);
				else getPathVec(G, u, G.n, prev, L, false);
				if (savedMove == 1) {
					doMove1(G, C, uPos, vPos, PathVec, newLen);
				}
				else if (savedMove == 2) {
					doMove2(G, C, uPos, vPos, PathVec, newLen);
				}
				else {
					doMove3(G, C, PathVec, newLen);
				}
			}
			//Update BSF if applicable			
			if (abs(C.len - G.k) < abs(CBest.len - G.k)) {
				CBest = C;
				CBest.improvedByOptimiser = true;
				if (CBest.len == G.k) {
					if (verbose >= 1) cout << "Cycle of length k = " << G.k << " found. Ending...\n";
					CBest.isGlobalOptimum = true;
					break;
				}
			}
			//Also update the global upper bound (for trimming purposes when doing s-cycles)
			if (C.len >= G.k && C.len < globalUB) {
				globalUB = C.len;
				if (doingSCycles && trimmingGraphs) {
					trimGraph(G, C, CBest, globalUB);
					if (G.n == G.m) {
						if (verbose >= 1) cout << "Graph has been trimmed so that n = m. Exiting optimisation process\n";
						return;
					}
				}
			}
		}
	}
	C = CBest;
}