﻿#include "ShortestPath.h"
#include "Graph.h"
#include "Connected.h"

vector<bool> D, considered, toConsider;
vector<int> col, S, inPathAtPos, P;
set<LabelItem, minLabel> PQ;
deque<int> intQ;

Sol getPathFromPrev(int s, int t, vector<int>& L, vector<int>& prev) {
	Sol P;
	P.len = L[t];
	if (P.len < infty) {
		int u = t;
		while (u != s) {
			P.seq.push_back(u);
			u = prev[u];
		}
		P.seq.push_back(s);
		reverse(P.seq.begin(), P.seq.end());
	}
	return P;
}
void DijkstraTarget(Graph& G, int s, int t, vector<int>& L, vector<int>& prev) {
	//Standard Dijkstra's algorithm from source s using a priority queue Q. Halts when (a) Q is empty, or (b) t becomes distinguished. 
	//If t is not in the vertex set (e.g. t == -1), then a shortest path tree rooted at s is returned in L and prev.
	//Any infty-weighted edges are ignored (i.e. we pretend they are not there)
	int u, v, w;
	PQ.clear();
	auto minPtr = PQ.begin();
	L.clear();
	L.resize(G.n);
	prev.clear();
	prev.resize(G.n);
	D.clear();
	D.resize(G.n);
	for (v = 0; v < G.n; v++) {
		D[v] = false;
		prev[v] = null;
		L[v] = infty;
	}
	L[s] = 0;
	PQ.emplace(LabelItem{ s, 0 });
	while (!PQ.empty()) {
		minPtr = PQ.begin();
		u = (*minPtr).vertex;
		PQ.erase(minPtr);
		D[u] = true;
		if (u == t) {
			return;
		}
		for (auto& e : G.AList[u]) {
			v = e.vertex;
			w = e.weight;
			if (w != infty && D[v] == false) {
				if (L[u] + w < L[v]) {
					if (L[v] != infty) {
						PQ.erase({ v, L[v] });
					}
					L[v] = L[u] + w;
					PQ.emplace(LabelItem{ v, L[v] });
					prev[v] = u;
				}
			}
		}
	}
}
void DijkstraSet(Graph& G, int s, set<int>& T, vector<int>& L, vector<int>& prev) {
	//Standard Dijkstra's algorithm from source s. Halts when (a) Q is empty, or (b) all vertices "toConsider" (in T) have become distinguished. 
	//Any infty-weighted edges are ignored (i.e. we pretend they are not there)
	int u, v, numConsidered = 0, w;
	PQ.clear();
	auto minPtr = PQ.begin();
	L.clear();
	L.resize(G.n);
	prev.clear();
	prev.resize(G.n);
	D.clear();
	D.resize(G.n);
	toConsider.clear();
	toConsider.resize(G.n);
	for (v = 0; v < G.n; v++) {
		D[v] = false;
		prev[v] = null;
		L[v] = infty;
		toConsider[v] = false;
	}
	for (auto& v : T) {
		toConsider[v] = true;
	}
	L[s] = 0;
	PQ.emplace(LabelItem{ s, 0 });
	while (!PQ.empty()) {
		minPtr = PQ.begin();
		u = (*minPtr).vertex;
		PQ.erase(minPtr);
		D[u] = true;
		if (toConsider[u] == true) {
			toConsider[u] = false;
			numConsidered++;
			if (numConsidered == T.size()) return;
		}
		for (auto& e : G.AList[u]) {
			v = e.vertex;
			w = e.weight;
			if (w != infty && D[v] == false) {
				if (L[u] + w < L[v]) {
					if (L[v] != infty) {
						PQ.erase({ v, L[v] });
					}
					L[v] = L[u] + w;
					PQ.emplace(LabelItem{ v, L[v] });
					prev[v] = u;
				}
			}
		}
	}
}
void DijkstraMaxDist(Graph& G, int s, int t, int dist, vector<int>& L, vector<int>& prev) {
	//Standard Dijkstra's algorithm from source s. Halts when (a) Q is empty, (b) all remaining undistinguished vertices are more than "dist" units from s, or
	//(c) when t becomes distinguished. Any infty-weighted edges are ignored (i.e. we pretend they are not there)
	int u, v, w;
	PQ.clear();
	auto minPtr = PQ.begin();
	L.clear();
	L.resize(G.n);
	prev.clear();
	prev.resize(G.n);
	D.clear();
	D.resize(G.n);
	for (v = 0; v < G.n; v++) {
		D[v] = false;
		prev[v] = null;
		L[v] = infty;
	}
	L[s] = 0;
	PQ.emplace(LabelItem{ s, 0 });
	while (!PQ.empty()) {
		minPtr = PQ.begin();
		if ((*minPtr).label > dist) {
			return;
		}
		u = (*minPtr).vertex;
		PQ.erase(minPtr);
		D[u] = true;
		if (u == t) {
			return;
		}
		for (auto& e : G.AList[u]) {
			v = e.vertex;
			w = e.weight;
			if (w != infty && D[v] == false) {
				if (L[u] + w < L[v]) {
					if (L[v] != infty) {
						PQ.erase({ v, L[v] });
					}
					L[v] = L[u] + w;
					PQ.emplace(LabelItem{ v, L[v] });
					prev[v] = u;
				}
			}
		}
	}
}
void randomBFSSet(Graph& G, int s, set<int>& T, vector<int>& L, vector<int>& prev) {
	//O(n+m) breadth-first search algorithm. Any infty-weighted edges are ignored (i.e. we pretend they are not there).
	//Also, neighbours of a vertex are considered in a random order using a copy of the graph called AVec
	//The process ends if all vertices in T have been painted black of if all vertices in G have been painted black. 
	int u, v, numConsidered = 0, w;
	L.clear();
	L.resize(G.n);
	prev.clear();
	prev.resize(G.n);
	col.clear();
	col.resize(G.n);
	toConsider.clear();
	toConsider.resize(G.n);
	for (v = 0; v < G.n; v++) {
		prev[v] = null;
		L[v] = infty;
		col[v] = white;
		toConsider[v] = false;
	}
	for (auto& v : T) {
		toConsider[v] = true;
	}
	//Make a shuffled version of the adjacency list to work with (O(n+m) operation)
	for (v = 0; v < G.n; v++) {
		G.AVec[v].clear();
		G.AVec[v].assign(G.AList[v].begin(), G.AList[v].end());
		random_shuffle(G.AVec[v].begin(), G.AVec[v].end());
	}
	intQ.clear();
	intQ.push_back(s);
	col[s] = grey;
	L[s] = 0;
	while (!intQ.empty()) {
		u = intQ.front();
		for (auto& el : G.AVec[u]) {
			v = el.vertex;
			w = el.weight;
			if (w != infty) {
				if (col[v] == white) {
					col[v] = grey;
					prev[v] = u;
					L[v] = L[u] + w;
					intQ.push_back(v);
				}
			}
		}
		intQ.pop_front();
		col[u] = black;
		if (toConsider[u] == true) {
			toConsider[u] = false;
			numConsidered++;
			if (numConsidered == T.size()) return;
		}
	}
}
void randomDFSSet(Graph& G, int s, set<int>& T, vector<int>& L, vector<int>& prev) {
	//O(n+m) depth-first search algorithm. Any infty-weighted edges are ignored (i.e. we pretend they are not there).
	//Also, neighbours of a vertex are considered in a random order using a copy of the graph called AVec
	//The process ends if all vertices in T have been painted black of if all vertices in G have been painted black. 
	int u, v, numConsidered = 0, w;
	L.clear();
	L.resize(G.n);
	prev.clear();
	prev.resize(G.n);
	col.clear();
	col.resize(G.n);
	toConsider.clear();
	toConsider.resize(G.n);
	for (v = 0; v < G.n; v++) {
		prev[v] = null;
		L[v] = infty;
		col[v] = white;
		toConsider[v] = false;
	}
	for (auto& v : T) {
		toConsider[v] = true;
	}
	//Make a shuffled version of the adjacency list to work with (O(n+m) operation)
	for (v = 0; v < G.n; v++) {
		G.AVec[v].clear();
		G.AVec[v].assign(G.AList[v].begin(), G.AList[v].end());
		random_shuffle(G.AVec[v].begin(), G.AVec[v].end());
	}
	S.clear();
	S.push_back(s);
	L[s] = 0;
	while (!S.empty()) {
		u = S.back();
		S.pop_back();
		if (col[u] != black) {
			col[u] = black;
			if (toConsider[u] == true) {
				toConsider[u] = false;
				numConsidered++;
				if (numConsidered == T.size()) return;
			}
			for (auto& el : G.AVec[u]) {
				v = el.vertex;
				w = el.weight;
				if (w != infty) {
					if (col[v] == white) {
						prev[v] = u;
						L[v] = L[u] + w;
						S.push_back(v);
					}
				}
			}
		}
	}
}
void RFSSet(Graph& G, int s, set<int>& T, vector<int>& L, vector<int>& prev) {
	//O(n+m) random-first search algorithm. Any infty-weighted edges are ignored (i.e. we pretend they are not there).
	//The process ends if all vertices in T have been painted black of if all vertices in G have been painted black. 
	int u, v, i, numConsidered=0, w;
	L.clear();
	L.resize(G.n);
	prev.clear();
	prev.resize(G.n);
	col.clear();
	col.resize(G.n);
	toConsider.clear();
	toConsider.resize(G.n);
	for (v = 0; v < G.n; v++) {
		prev[v] = null;
		L[v] = infty;
		col[v] = white;
		toConsider[v] = false;
	}
	for (auto& v : T) {
		toConsider[v] = true;
	}
	S.clear();
	S.push_back(s);
	L[s] = 0;
	while (!S.empty()) {
		//Choose random vertex in S
		i = rand() % S.size();
		u = S[i];
		swapVals(S[i], S.back());
		S.pop_back();
		if (col[u] != black) {
			col[u] = black;
			if (toConsider[u] == true) {
				toConsider[u] = false;
				numConsidered++;
				if (numConsidered == T.size()) return;
			}
			for (auto& el : G.AList[u]) {
				v = el.vertex;
				w = el.weight;
				if (w != infty) {
					if (col[v] == white) {
						prev[v] = u;
						L[v] = L[u] + w;
						S.push_back(v);
					}
				}
			}
		}
	}
}
void generateLoopFreePath(Graph &G) {
	int u, v, i;
	inPathAtPos.clear();
	inPathAtPos.resize(G.n, null);
	inPathAtPos[P.back()] = 0;
	while (true) {
		u = P.back();
		i = rand() % G.AVec[u].size();
		v = G.AVec[u][i].vertex;
		if (considered[v]) {
			//Have generated a path back to the tree so end
			P.push_back(v);
			return;
		}
		else if (inPathAtPos[v] == null) {
			//Extending the path by one vertex
			P.push_back(v);
			inPathAtPos[v] = P.size() - 1;
		}
		else {
			//Creating a cycle at v so backtrack the path to first occurence of v
			while (P.size() - 1 > inPathAtPos[v]) {
				inPathAtPos[P.back()] = null;
				P.pop_back();
			}
		}
	}
}
set<int> getReachableVertices(Graph& G, int s) {
	//Use BFS to produce the set of vertices reachable from the vertex s
	int u, v;
	col.clear();
	col.resize(G.n, white);
	intQ.clear();
	col[s] = grey;
	intQ.push_back(s);
	while (!intQ.empty()) {
		u = intQ.front();
		for (auto& el : G.AList[u]) {
			v = el.vertex;
			if (el.weight != infty && col[v] == white) {
				col[v] = grey;
				intQ.push_back(v);
			}
		}
		intQ.pop_front();
		col[u] = black;
	}
	//All vertices reachable from s (including s) are coloured black.
	set<int> reachable;
	for (v = 0; v < G.n; v++) {
		if (col[v] == black) reachable.insert(v);
	}
	return(reachable);
}
void WilsonSet(Graph& G, int s, set<int>& T, vector<int>& L, vector<int>& prev) {
	//O(exp) Wilson's algorithm for producing a spanning tree randomly.
	int u, v, i, sPrime = G.n - 1;
	L.clear();
	L.resize(G.n);
	prev.clear();
	prev.resize(G.n);
	considered.clear();
	considered.resize(G.n);
	P.clear();
	//Because the current G can be disconnected, we need to ID all vertices in G reachable from s. We will call this set of vertices X. 
	set<int> X = getReachableVertices(G, s);
	//Now make a vectorised version of the adjacency list of G (O(n+m) operation). This will allow fast random selection of neighbours
	//for use in Wilson's algorithm. Note that we don't include the infinty weighted edges in G and we also reverse all edges. The second 
	//bit is done because Wilson's algorithm operates by finding paths from each v in X to s (not the other way round).
	for (u = 0; u < G.n; u++) {
		G.AVec[u].clear();
	}
	for (u = 0; u < G.n; u++) {
		for (auto& el : G.AList[u]) {
			if (el.weight != infty) {
				G.AVec[el.vertex].push_back({ u, el.weight });
			}
		}
	}
	//Initialise the other arrays. A vertex has been "considered" if (a) it is not connected to s in G, or (b) it has
	//already been put into the Wilson tree we're constructing
	for (v = 0; v < G.n; v++) {
		prev[v] = null;
		L[v] = infty;
		considered[v] = true;
	}
	for (auto& v : X) considered[v] = false;
	//Now run Wilson's algorithm
	considered[s] = true;
	L[s] = 0;
	X.erase(s);
	while (!X.empty()) {
		//Choose a random vertex u that is not yet considered and determine a path to a vertex in the tree
		auto it = X.begin();
		advance(it, rand() % X.size());
		u = *it;
		P = { u };
		generateLoopFreePath(G);
		//Reverse the path so that it goes from the tree to u
		reverse(P.begin(), P.end());
		//Go through this path, add it to the tree, mark its vertices as considered, and calculate the lengths from the source
		for (i = 1; i < P.size(); i++) {
			prev[P[i]] = P[i - 1];
			L[P[i]] = L[P[i - 1]] + getArcWeight(G, P[i - 1], P[i]);
			considered[P[i]] = true;
			X.erase(P[i]);
		}
	}
}
