#include "main.h"
#include "Graph.h"
#include "IO.h"
#include "LocalSearch.h"
#include "ShortestPath.h"
#include "GraphStats.h"
#include "Yen.h"
#include "IP.h"

int verbose;

int minVal(int a, int b) {
	if (a < b) return a;
	else return b;
}
void swapVals(int& a, int& b) {
	int x = a; a = b; b = x;
}
double stDev(vector<int>& X) {
	double sumX = 0.0, sumXSq = 0.0, n = (double)X.size(), x;
	for (auto& el: X) {
		x = (double)el;
		sumX += x;
		sumXSq += pow(x, 2.0);
	}
	return(sqrt((sumXSq - (pow(sumX, 2) / n)) / (n - 1)));
}
bool fileExists(string& name) {
	ifstream file(name);
	if (!file) return false;
	else return true;
}
void logAndExit(string s) {
	//Writes message s to screen and log file and then exits the program
	string fileName = "resLog.txt";
	if (!fileExists(fileName)) {
		ofstream resultsLog("resLog.txt");
		resultsLog << "inFile\tk\tTimeLimt\tStartNode\tSPAlg\tOptimisationAlg\tTrimmingGraphs\tSeed\tVerbose\tBestLen\tImprovedByOptimiser\tProvenGlobalOpt\tOverallRunTime\tBestCycleSeen\n";
		resultsLog.close();
	}
	ofstream resultsLog("resLog.txt", ios::app);
	resultsLog << s;
	cout << s;
	resultsLog.close();
	exit(1);
}
void usage() {
	cout << "Optimisation algorithm for the k-length cycle problem (KCP) and the k-length s-cycle problem (KSCP) in edge-weighted undirected graphs.\n"
		<< "(R. Lewis, Cardiff University, 2023. www.rhydlewis.eu)\n"
		<< "Usage:\n"
		<< "---\n"
		<< "-i <string>       (Required. Must define an input graph in the recognised format. See the documentation for further details.)\n"
		<< "---\n"
		<< "-k <int>          (Desired length of cycle (default = 1000). If -k is given a negative value the desired length is set to infinity and the algorithm seeks the longest cycle)\n"
		<< "-s <int>          (Index for the source vertex for KSCP. If a negative value is specified, there is no source vertex and the KCP is considered (default = -1).)\n"
		<< "-t <double>       (Algorithm time limit in seconds (default = 5).)\n"
		<< "-A <int>          (Optimisation algorithm to use (default = 1):\n"
		<< "                         1: Great deluge (local search) algorithm\n"
		<< "                         2: Yen's algorithm (requires Yen.Jar and a local Java installation)\n"
		<< "                         3: Integer programming (requires k-cycle.mos, k-s-cycle.mos, and local installation of Fico Xpress)\n"
		<< "-a <int>          (Path-finding algorithm used with local search (default = 1):\n"
		<< "                         1: Random first search\n"
		<< "                         2: Breadth first search (randomised)\n"
		<< "                         3: Depth first search (randomised)\n"
		<< "                         4: Wilson's algorithm\n"
		<< "                         5: Dijkstra's algorithm)\n"
		<< "-T                (Only has an effect with the KSCP. If -T is present, the graph is not reduced in size during a run. (When -T is not present, whenever the best solution\n"
		<< "                   whose cost > k is improved, graphs are trimmed by removing all vertices whose distance from the source is greater or equal to floor(cost / 2) + 1.)\n"
		<< "-r <int>          (Random seed (default = 1).)\n"
		<< "-v                (Verbosity. If present, output is sent to the console. If -v is repeated, more output is given.)\n"
		<< "-stats            (If present, no optimisation is performed. Instead, stats on each BCC are just written to the screen.)\n"
		<< "---\n";
	exit(1);
}
int main(int argc, char** argv) {

	//Convert the input arguments to a vector of strings
	vector<string> argList(argv, argv + argc);
	if (argList.size() <= 1) usage();

	//Main variables
	int i, startNode = -1, targetK = 1000, seed = 1, SPAlg = 1, globalUB = infty, optimisationAlg = 1;
	double timeLimit = 5.0, endTemp = 0.01;
	bool doingSCycles = false, printStats = false, trimmingGraphs = true;
	vector<set<Neighbour, minNeighbour> > adjList;
	vector<string> vName;
	vector<Graph> subGs;
	string inFile = "";
	verbose = 0;

	//Parse the command line arguments
	try {
		for (i = 1; i < argList.size(); i++) {
			if (argList[i] == "-s") {
				startNode = stoi(argList[++i]);
				if (startNode >= 0) doingSCycles = true;
			}
			else if (argList[i] == "-k") targetK = stoi(argList[++i]);
			else if (argList[i] == "-t") timeLimit = stof(argList[++i]);
			else if (argList[i] == "-a") SPAlg = stoi(argList[++i]);
			else if (argList[i] == "-r") seed = stoi(argList[++i]);
			else if (argList[i] == "-e") endTemp = stof(argList[++i]);
			else if (argList[i] == "-T") trimmingGraphs = false;
			else if (argList[i] == "-v") verbose++;
			else if (argList[i] == "-stats") printStats = true;
			else if (argList[i] == "-A") optimisationAlg = stoi(argList[++i]);
			else if (argList[i] == "-i") {
				inFile = argList[++i];
				readInputFile(inFile, adjList, vName);
			}
			else {
				cout << "Error with input statement. Ending...\n";
				usage();
			}
		}
	}
	catch (...) {
		cout << "Error with input statement. Ending...\n";
		usage();
	}

	//Check some input combinations
	if (startNode >= (int)adjList.size()) logAndExit("Invalid algorithm option. Specified start vertex does not exist. Ending...\n");
	if (!doingSCycles && optimisationAlg == 2) logAndExit("Invalid algorithm option. Yen's algorithm must be run from a specific source vertex. Ending...\n");
	if (optimisationAlg < 1 || optimisationAlg > 3) logAndExit("Invalid optimisation algorithm selected. Must be in {1,2,3}. Ending...\n");
	if (!doingSCycles && trimmingGraphs) trimmingGraphs = false;
	if (targetK < 0) targetK = infty;

	//Set random seed
	srand(seed);

	//Set output file to null
	ofstream fout(inFile + "_sol.txt");
	fout << "null\n";
	fout.close();

	//Start the timer and calculate when the run is due to end
	clock_t runStart = clock();
	clock_t runEnd = runStart + (clock_t)(timeLimit * CLOCKS_PER_SEC);

	//Read in the graph from the input file, then determine the set of biconnected components that are non dyads and put these into the vector subGs (sub-graphs).
	//The k-values for each subgraph G_i are set at min(targetK, UB(G_i)) where UB(G_i) is the sum of the n longest edges in G_i.
	{
		Graph G = makeGraph(adjList, vName, targetK, startNode);
		vector<set<int>> BCCs = getBCCs(G);
		for (i = 0; i < BCCs.size(); i++) {
			if (BCCs[i].size() >= 3) {
				if (!doingSCycles) {
					subGs.push_back(getInducedGraph(G, BCCs[i]));
				}
				else if (BCCs[i].count(G.source) == 1) {
					subGs.push_back(getInducedGraph(G, BCCs[i]));
				}
			}
		}
		//Finally the non-dyad biconnected components are sorted by the number of edges (largest first)
		sort(subGs.begin(), subGs.end());
		reverse(subGs.begin(), subGs.end());
		if (subGs.empty()) logAndExit("Note: Considered is a tree / forest. There is no available cycle. Ending...\n");
	}

	if (printStats) {
		cout << "InputGraph:         " << inFile << endl;
		cout << "StartNode:          " << startNode << endl;
		cout << "NumNonDyadBCCs:     " << subGs.size() << endl;
		for (i = 0; i < subGs.size(); i++) {
			cout << "  BCC-" << i << endl;
			GraphStats GS = getGraphStats(subGs[i]);
			printGraphStats(GS);
		}
		logAndExit("Stats run on " + inFile + "\n");
	}

	//Produce a starting solution for each biconnected component in G and keep track of the best solution seen
	vector<Sol> cycles, candidates;
	Sol bestSeen;
	int bestSubG = -1;
	for (i = 0; i < subGs.size(); i++) {
		//For each biconnected component produce up to 4 candidate solutions using the 4 (polynomial time) path-finding algorithms
		candidates.push_back(getInitSol(subGs[i], 1));
		if (subGs[i].n != subGs[i].m) {
			//BCC Graph is not a cycle, so produce some other solutions too
			candidates.push_back(getInitSol(subGs[i], 2));
			candidates.push_back(getInitSol(subGs[i], 3));
			candidates.push_back(getInitSol(subGs[i], 5));
		}
		else {
			candidates.back().isGlobalOptimum = true;
		}
		//Save the best of these as the initial solution to this BCC. If we are considering s-cycles, also keep tabs on the length 
		//of the best observed solution that is longer than the targetK (globalUB). This will allow us to trim the graphs when optimising with s-cycles
		int min = infty, minPos = null;
		for (int j = 0; j < candidates.size(); j++) {
			if (abs(candidates[j].len - subGs[i].k) < min) {
				min = abs(candidates[j].len - subGs[i].k);
				minPos = j;
			}
			if (candidates[j].len >= targetK && candidates[j].len < globalUB) {
				globalUB = candidates[j].len;
			}
		}
		cycles.push_back(candidates[minPos]);
		if (cycles.back().len == subGs[i].k) cycles.back().isGlobalOptimum = true;
		if (abs(cycles.back().len - targetK) < abs(bestSeen.len - targetK)) {
			bestSeen = cycles.back();
			bestSubG = i;
			if (bestSeen.len == targetK) break;
		}
		candidates.clear();
	}

	//If desired, print some stats on each biconnected component and its inital solution.
	if (verbose >= 1) {
		cout << "InputGraph (with source " << startNode << ") comprises " << subGs.size() << " biconnected componets of the following (n, m, k) details:\n";
		for (i = 0; i < subGs.size(); i++) {
			cout << "G_" << i << " = (" << subGs[i].n << ", " << subGs[i].m << ", " << subGs[i].k << "). Initial solution has length " << cycles[i].len <<"\n";
		}
	}

	//If a k-length solution has not been found, decide which solutions need to be optimised and allocate them time.
	if (bestSeen.len != targetK) {
		int totalProbSize = 0;
		for (i = 0; i < subGs.size(); i++) {
			if (cycles[i].isGlobalOptimum == false) {
				totalProbSize += subGs[i].n + subGs[i].m;
			}
		}
		//Now optimise the solution belonging to each biconnected component according to a time limit
		clock_t ticsRemaining = (clock_t)runEnd - clock();
		for (i = 0; i < subGs.size(); i++) {
			if (cycles[i].isGlobalOptimum == false) {
				clock_t clockLimit = clock() + (clock_t)(ticsRemaining * ((subGs[i].n + subGs[i].m) / double(totalProbSize)));
				if (subGs[i].LenUB < targetK && targetK - subGs[i].LenUB >= abs(targetK - bestSeen.len)) {
					if (verbose >= 1) cout << "---\nSkipping optimisation phase on subgraph G_" << i << ". Its UB of " << subGs[i].LenUB << " is too far below the target length\n";
				}
				else {
					if (optimisationAlg == 1) {
						if (verbose >= 1) cout << "---\nRunning GD on subgraph G_" << i << ", k = " << subGs[i].k << " for " << (clockLimit - clock()) / double(CLOCKS_PER_SEC) << "s. Best solution so far (BSF) has length " << bestSeen.len << "\n";
						doGreatDeluge(subGs[i], cycles[i], SPAlg, clockLimit, doingSCycles, trimmingGraphs, globalUB);
					}
					else if (optimisationAlg == 2) {
						if (verbose >= 1) cout << "---\nRunning Yen's algorithm on subgraph G_" << i << ", k = " << subGs[i].k << " for " << (clockLimit - clock()) / double(CLOCKS_PER_SEC) << "s. Best solution so far (BSF) has length " << bestSeen.len << "\n";
						doYen(subGs[i], cycles[i], (clockLimit - clock()) / double(CLOCKS_PER_SEC));
					}
					else {
						if (verbose >= 1) cout << "---\nRunning IP on subgraph G_" << i << ", k = " << subGs[i].k << " for " << (clockLimit - clock()) / double(CLOCKS_PER_SEC) << "s. Best solution so far (BSF) has length " << bestSeen.len << "\n";
						doIP(subGs[i], cycles[i], (clockLimit - clock()) / double(CLOCKS_PER_SEC), doingSCycles);
					}
					if (abs(cycles[i].len - targetK) <= abs(bestSeen.len - targetK)) {
						bestSeen = cycles[i];
						bestSubG = i;
						if (bestSeen.len == targetK) break;
					}
				}
			}
		}
	}	

	//Write the best solution to a file
	printSolAndCheck(subGs[bestSubG], bestSeen, false);
	writeSolToFile(subGs[bestSubG], bestSeen, inFile);
	double overallRunTime = (clock() - runStart) / (double)(CLOCKS_PER_SEC);
	if (verbose >= 1) cout << "Total time taken in run = " << overallRunTime << " s." << endl;

	//Output some information (on a single line) to the screen and also append it to the log file 'resLog.txt'
	string s =
		inFile + "\t" +
		to_string(targetK) + "\t" +
		to_string(timeLimit) + "\t" +
		to_string(startNode) + "\t" +
		to_string(SPAlg) + "\t" +
		to_string(optimisationAlg) + "\t" +
		to_string(trimmingGraphs) + "\t" +
		to_string(seed) + "\t" +
		to_string(verbose) + "\t" +
		to_string(bestSeen.len) + "\t" +
		to_string(bestSeen.improvedByOptimiser) + "\t" +
		to_string(bestSeen.isGlobalOptimum) + "\t" +
		to_string(overallRunTime) + "\t" +
		prettySolStringOrigLabels(subGs[bestSubG], bestSeen) + "\n";
	logAndExit(s);
	
	return 0;
}